#!/usr/bin/env python

"""
Common scheduling functionality.

Copyright (C) 2015, 2016 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from imiptools.text import parse_line
from imiptools.handlers.scheduling.manifest import confirmation_functions, \
                                                   locking_functions, \
                                                   retraction_functions, \
                                                   scheduling_functions, \
                                                   unlocking_functions

# Function application/invocation.

def apply_scheduling_functions(functions, handler):

    """
    Apply the given scheduling 'functions' in the current object of the given
    'handler'. This function starts a transaction that should be finalised using
    the 'finish_scheduling' function.
    """

    # First, lock the resources to be used.

    start_scheduling(functions, handler)

    # Obtain the actual scheduling functions with arguments.

    schedulers = get_function_calls(functions, scheduling_functions)

    # Then, invoke the scheduling functions.

    response = "ACCEPTED"

    for fn, args in schedulers:

        # NOTE: Should signal an error for incorrectly configured resources.

        if not fn:
            return "DECLINED"

        # Keep evaluating scheduling functions, stopping only if one
        # declines or gives a null response.

        else:
            result = fn(handler, args)

            # Return a negative result immediately.

            if not result or result == "DECLINED":
                return result

            # Modify the eventual response from acceptance if a countering
            # result is obtained.

            elif response == "ACCEPTED":
                response = result

    return response

def confirm_scheduling(functions, handler):

    """
    Confirm scheduling using the given listener 'functions' for the current
    object of the given 'handler'. This function continues a transaction that
    should be finalised using the 'finish_scheduling' function.
    """

    # Obtain the actual listener functions with arguments.

    functions = get_function_calls(functions, confirmation_functions)
    apply_functions(functions, handler)

def retract_scheduling(functions, handler):

    """
    Retract scheduling using the given listener 'functions' for the current
    object of the given 'handler'. This function is a complete transaction in
    itself.
    """

    # First, lock the resources to be used.

    start_scheduling(functions, handler)

    # Obtain the actual listener functions with arguments.

    retractors = get_function_calls(functions, retraction_functions)
    apply_functions(retractors, handler)

    # Finally, unlock the resources.

    finish_scheduling(functions, handler)

def start_scheduling(functions, handler):

    """
    Apply locking functions for the given scheduling 'functions' and for the
    current object of the given 'handler'.
    """

    # Obtain functions to lock resources.

    locks = get_function_calls(functions, locking_functions)
    apply_functions(locks, handler)

def finish_scheduling(functions, handler):

    """
    Finish scheduling using the given scheduling 'functions' for the current
    object of the given 'handler'.
    """

    # Obtain functions to unlock resources.

    locks = get_function_calls(functions, unlocking_functions)
    apply_functions(locks, handler)

def apply_functions(functions, handler):

    """
    Apply the given notification 'functions' for the current object of the given
    'handler'.
    """

    for fn, args in functions:

        # NOTE: Should signal an error for incorrectly configured resources.

        if not fn:
            continue

        fn(handler, args)

# Function lookup.

def get_function_calls(lines, registry):

    """
    Parse the given 'lines', returning a list of (function, arguments) tuples,
    with each function being a genuine function object and with the arguments
    being a list of strings.

    Each of the 'lines' should employ the function name and argument strings
    separated by whitespace, with any whitespace inside arguments quoted using
    single or double quotes.

    The given 'registry' indicates the mapping from function names to actual
    functions.
    """

    functions = []

    for line in lines:
        parts = parse_line(line)
        functions.append((registry.get(parts[0]), parts[1:]))

    return functions

# vim: tabstop=4 expandtab shiftwidth=4
